import json
import os
from datetime import datetime
from decimal import Decimal
from unittest import mock

from satcfdi.create import Issuer
from satcfdi.pacs import Environment
from satcfdi.pacs.diverza import Diverza
from satcfdi.create.cfd import cfdi33, cfdi40, pago20
from tests.utils import get_signer, verify_result

current_dir = os.path.dirname(__file__)
module = 'satcfdi'


def test_diverza_validate_rfc():
    diverza = Diverza(
        rfc="AAA010101AAA",
        id="3935",
        token="ABCD1234"
    )

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={'identifiers': [{'rfc': 'AAA010101AAA', 'active': 'Y'}]})
        
        r = diverza.rfc_valid("AAA010101AAA")
        assert r

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={'identifiers': [{'rfc': 'AAA010101AAA', 'active': 'Y'}, {'rfc': 'ABC010101DAS', 'active': 'N'}]})
        
        r = diverza.rfc_valid(["AAA010101AAA", "ABC010101DAS"])

        assert r[0]
        assert not r[1]


def test_diverza_stamp():
    signer = get_signer('xiqb891116qe4')
    emisor = Issuer(signer=signer, tax_system="601")

    diverza = Diverza(
        rfc="AAA010101AAA",
        id="3935",
        token="ABCD1234",
        environment=Environment.TEST
    )

    cfdi = cfdi33.Comprobante(
        emisor=emisor,
        lugar_expedicion="56820",
        fecha=datetime.fromisoformat("2022-09-28T22:40:38"),
        receptor=cfdi33.Receptor(
            rfc='KIJ0906199R1',
            nombre='KIJ, S.A DE C.V.',
            uso_cfdi='G03'
        ),
        metodo_pago='PPD',
        serie="A",
        folio="123456",
        conceptos=[
            cfdi33.Concepto(
                cuenta_predial='1234567890',
                clave_prod_serv='10101702',
                cantidad=Decimal('1.00'),
                clave_unidad='E48',
                descripcion='SERVICIOS DE FACTURACION',
                valor_unitario=Decimal('15390.30'),
                impuestos=cfdi33.Impuestos(
                    traslados=cfdi33.Impuesto.parse('002|Tasa|0.160000'),
                    retenciones=[cfdi33.Impuesto.parse('001|Tasa|0.100000'), cfdi33.Impuesto.parse('002|Tasa|0.106667')],
                )
            )
        ]
    )

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={
            "uuid": "3d94d0c5-9d4d-4c73-a61b-d083c553b7cc",
            "content": "dGhpcyBpcyBhIHRlc3QgeG1s"
        })
        diverza.stamp(cfdi=cfdi)

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2)
        verify = verify_result(data=args, filename="test_diverza_stamp.json")
        assert verify

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={
            "uuid": "3d94d0c5-9d4d-4c73-a61b-d083c553b7cc",
            "content": "dGhpcyBpcyBhIHRlc3QgeG1s"
        })
        diverza.issue(
            cfdi=cfdi,
            # certificate_number=signer.certificate_number()
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2)
        verify = verify_result(data=args, filename="test_diverza_issue.json")
        assert verify


def test_diverza_stamp_v40():
    signer = get_signer('xiqb891116qe4')
    emisor = Issuer(signer=signer, tax_system="601")

    diverza = Diverza(
        rfc="AAA010101AAA",
        id="3935",
        token="ABCD1234",
        environment=Environment.TEST
    )

    cfdi = cfdi40.Comprobante(
        emisor=emisor,
        lugar_expedicion="56820",
        fecha=datetime.fromisoformat("2022-09-28T22:40:38"),
        receptor=cfdi40.Receptor(
            rfc='KIJ0906199R1',
            nombre='KIJ, S.A DE C.V.',
            uso_cfdi='G03',
            domicilio_fiscal_receptor="59820",
            regimen_fiscal_receptor="601"
        ),
        metodo_pago='PPD',
        serie="A",
        folio="123456",
        conceptos=[
            cfdi40.Concepto(
                cuenta_predial='1234567890',
                clave_prod_serv='10101702',
                cantidad=Decimal('1.00'),
                clave_unidad='E48',
                descripcion='SERVICIOS DE FACTURACION',
                valor_unitario=Decimal('15390.30'),
                impuestos=cfdi40.Impuestos(
                    traslados=cfdi40.Impuesto.parse('002|Tasa|0.160000'),
                    retenciones=[cfdi40.Impuesto.parse('001|Tasa|0.100000'), cfdi40.Impuesto.parse('002|Tasa|0.106667')],
                )
            )
        ]
    )

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={
            "uuid": "3d94d0c5-9d4d-4c73-a61b-d083c553b7cc",
            "content": "dGhpcyBpcyBhIHRlc3QgeG1s"
        })
        diverza.stamp(cfdi=cfdi)

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2)
        verify = verify_result(data=args, filename="test_diverza_stamp_v40.json")
        assert verify

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={
            "uuid": "3d94d0c5-9d4d-4c73-a61b-d083c553b7cc",
            "content": "dGhpcyBpcyBhIHRlc3QgeG1s"
        })
        diverza.issue(
            cfdi=cfdi,
            # certificate_number=signer.certificate_number()
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2)
        verify = verify_result(data=args, filename="test_diverza_issue_v40.json")
        assert verify


def test_diverza_stamp_pago_v40():
    signer = get_signer('xiqb891116qe4')
    emisor = Issuer(signer=signer, tax_system="601")

    diverza = Diverza(
        rfc="AAA010101AAA",
        id="3935",
        token="ABCD1234",
        environment=Environment.TEST
    )

    cfdi = cfdi40.Comprobante.pago(
        emisor=emisor,
        fecha=datetime.fromisoformat("2022-09-28T22:40:38"),
        receptor=cfdi40.Receptor(
            rfc='KIJ0906199R1',
            nombre='KIJ, S.A DE C.V.',
            uso_cfdi='G03',
            domicilio_fiscal_receptor="59820",
            regimen_fiscal_receptor="601"
        ),
        lugar_expedicion="56820",
        complemento_pago=pago20.Pagos(
            pago=pago20.Pago(
                fecha_pago=datetime(2020, 1, 1),
                forma_de_pago_p='03',
                moneda_p='MXN',
                tipo_cambio_p=1,
                docto_relacionado=pago20.DoctoRelacionado(
                    id_documento='d6042dc8-d525-4e78-8d1b-092c878bd518',
                    imp_pagado=Decimal("100.3"),
                    imp_saldo_ant=Decimal("203.45"),
                    num_parcialidad=3,
                    moneda_dr="MXN",
                    objeto_imp_dr="01"
                )
            )
        ),
        serie="A",
        folio="123456"
    )

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={
            "uuid": "3d94d0c5-9d4d-4c73-a61b-d083c553b7cc",
            "content": "dGhpcyBpcyBhIHRlc3QgeG1s"
        })
        diverza.stamp(cfdi=cfdi)

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2)
        verify = verify_result(data=args, filename="test_diverza_stamp_pago_v40.json")
        assert verify

    with mock.patch(f'requests.request') as mk:
        mk.return_value.json = mock.Mock(return_value={
            "uuid": "3d94d0c5-9d4d-4c73-a61b-d083c553b7cc",
            "content": "dGhpcyBpcyBhIHRlc3QgeG1s"
        })
        diverza.issue(
            cfdi=cfdi,
            # certificate_number=signer.certificate_number()
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'

        args = json.dumps(mk.call_args.kwargs, indent=2)
        verify = verify_result(data=args, filename="test_diverza_issue_pago_v40.json")
        assert verify




"""ecc11 http://www.sat.gob.mx/EstadoDeCuentaCombustible"""
from decimal import Decimal
from datetime import datetime, date, time
from collections.abc import Sequence
from ... import CFDI, XElement, ScalarMap


class Traslado(ScalarMap):
    """
    Nodo para la definición de información detallada de un traslado de impuesto específico
    
    :param impuesto: Atributo requerido para definir el tipo de impuesto trasladado.
    :param tasao_cuota: Atributo requerido para señalar la tasa o la cuota del impuesto que se traslada por cada concepto amparado en el comprobante. Cuando se registre un porcentaje, por ejemplo 16%, debe expresarse como 0.16 y no como 16.00
    :param importe: Atributo requerido para definir el importe o monto del impuesto trasladado.
    """
    
    def __init__(
            self,
            impuesto: str,
            tasao_cuota: Decimal | int,
            importe: Decimal | int,
    ): 
        super().__init__({
            'Impuesto': impuesto,
            'TasaoCuota': tasao_cuota,
            'Importe': importe,
        })
        

class ConceptoEstadoDeCuentaCombustible(ScalarMap):
    """
    Nodo requerido para la expresión de una transacción a ser reportada en el estado de cuenta del proveedor de monedero electrónico para operaciones de compra de combustibles.
    
    :param identificador: Atributo requerido para la expresión del identificador o número del monedero electrónico.
    :param fecha: Atributo requerido para la expresión de la Fecha y hora de expedición de la operación reportada. Se expresa en la forma aaaa-mm-ddThh:mm:ss, de acuerdo con la especificación ISO 8601.
    :param rfc: Atributo requerido del RFC del enajenante del combustible.
    :param clave_estacion: Atributo requerido para expresar la clave de cliente de la estación de servicio, a 10 caracteres.
    :param cantidad: Atributo requerido para definir el volumen de combustible adquirido.
    :param no_identificacion: Atributo requerido para indicar la clave del producto del combustible.
    :param nombre_combustible: Atributo requerido para expresar el nombre del combustible adquirido.
    :param folio_operacion: Atributo requerido para referir el número de folio de cada operación realizada por cada monedero electrónico.
    :param valor_unitario: Atributo requerido para definir el precio unitario del combustible adquirido.
    :param importe: Atributo requerido para definir el monto total de consumo de combustible. Debe ser equivalente al resultado de multiplicar la cantidad por el valor unitario, redondeado a centésimas.
    :param traslados: Nodo requerido para enlistar los impuestos trasladados aplicables de combustibles.
    :param tar: Atributo condicional para expresar la clave de la Terminal de Almacenamiento y Reparto (CVE TAR), conforme al catálogo publicado en la página de Internet del SAT, mismo que servirá para identificar la cuota por litro conforme a las tablas que publique la Secretaría de Hacienda y Crédito Público para determinar el monto del estímulo fiscal.
    :param unidad: Atributo condicional para precisar la unidad de medida.
    """
    
    def __init__(
            self,
            identificador: str,
            fecha: datetime,
            rfc: str,
            clave_estacion: str,
            cantidad: Decimal | int,
            no_identificacion: str,
            nombre_combustible: str,
            folio_operacion: str,
            valor_unitario: Decimal | int,
            importe: Decimal | int,
            traslados: Traslado | dict | Sequence[Traslado | dict],
            tar: str = None,
            unidad: str = None,
    ): 
        super().__init__({
            'Identificador': identificador,
            'Fecha': fecha,
            'Rfc': rfc,
            'ClaveEstacion': clave_estacion,
            'Cantidad': cantidad,
            'NoIdentificacion': no_identificacion,
            'NombreCombustible': nombre_combustible,
            'FolioOperacion': folio_operacion,
            'ValorUnitario': valor_unitario,
            'Importe': importe,
            'Traslados': traslados,
            'TAR': tar,
            'Unidad': unidad,
        })
        

class EstadoDeCuentaCombustible(CFDI):
    """
    Complemento para el Comprobante Fiscal Digital por Internet (CFDI) para integrar la información aplicable al estado de cuenta emitido por un prestador de servicios de monedero electrónico
    
    :param numero_de_cuenta: Atributo requerido para expresar el número de cuenta del adquirente del monedero electrónico
    :param sub_total: Atributo requerido para representar la suma de todos los importes tipo ConceptoEstadoDeCuentaCombustible.
    :param total: Atributo requerido para expresar el monto total de consumo de combustible.
    :param conceptos: Nodo requerido para enlistar los conceptos cubiertos por Estado de Cuenta de Combustible.
    """
    
    tag = '{http://www.sat.gob.mx/EstadoDeCuentaCombustible}EstadoDeCuentaCombustible'
    version = '1.1'
    tipo_operacion = 'Tarjeta'
    
    def __init__(
            self,
            numero_de_cuenta: str,
            sub_total: Decimal | int,
            total: Decimal | int,
            conceptos: ConceptoEstadoDeCuentaCombustible | dict | Sequence[ConceptoEstadoDeCuentaCombustible | dict],
    ): 
        super().__init__({
            'Version': self.version,
            'TipoOperacion': self.tipo_operacion,
            'NumeroDeCuenta': numero_de_cuenta,
            'SubTotal': sub_total,
            'Total': total,
            'Conceptos': conceptos,
        })
        




import json
from datetime import datetime
from decimal import Decimal
from unittest import mock

from satcfdi.create import Issuer
from satcfdi.create.cfd import cfdi40
from satcfdi.pacs import Environment
from satcfdi.pacs.comerciodigital import ComercioDigital
from utils import get_signer, verify_result


def test_comercio_digital_test():
    comercio_digital = ComercioDigital(
        user="AAA010101AAA",
        password="PWD",
        environment=Environment.TEST
    )

    signer = get_signer('xia190128j61')
    emisor = Issuer(signer=signer, tax_system="601")

    invoice = cfdi40.Comprobante(
        emisor=emisor,
        lugar_expedicion="27200",
        fecha=datetime.fromisoformat("2022-09-28T22:40:38"),
        receptor=cfdi40.Receptor(
            rfc="ARE100609JP4",
            nombre="ACE RENUEVA SA DE CV",
            uso_cfdi="G03",
            domicilio_fiscal_receptor="85000",
            regimen_fiscal_receptor="601"
        ),
        metodo_pago="PPD",
        forma_pago="99",
        serie="T",
        folio="1000",
        conceptos=cfdi40.Concepto(
            cuenta_predial='1234567890',
            clave_prod_serv='10101702',
            cantidad=Decimal('.10'),
            clave_unidad='E48',
            descripcion='SERVICIOS DE RENTA',
            valor_unitario=Decimal("1.00"),
            impuestos=cfdi40.Impuestos(
                traslados=cfdi40.Impuesto.parse('002|Tasa|0.160000'),
                retenciones=[cfdi40.Impuesto.parse('001|Tasa|0.100000'), cfdi40.Impuesto.parse('002|Tasa|0.106667')],
            ),
            _traslados_incluidos=True
        )
    )

    with mock.patch(f'requests.request') as mk:
        mk.return_value.ok = True
        mk.return_value.headers = {
            'codigo': '000',
            'uuid': '66ea0f5a-4f00-4406-a573-87eb1a34f1fa'
        }
        mk.return_value.content = b'Hello'

        comercio_digital.stamp(
            cfdi=invoice
        )

        assert mk.called
        mk.call_args.kwargs["headers"]["User-Agent"] = 'this is a test'
        mk.call_args.kwargs["data"] = b"".join(mk.call_args.kwargs["data"])

        args = json.dumps(mk.call_args.kwargs, indent=2, default=str, ensure_ascii=False)
        verify = verify_result(data=args, filename="test_stamp.json")
        assert verify

